import numpy as np
from agox.postprocessors.ABC_postprocess import PostprocessBaseClass
from agox.utils.ray import RayPoolUser, Task
from ase.calculators.calculator import all_properties
from ase.calculators.singlepoint import SinglePointCalculator
from ase.optimize import BFGS
from ase.constraints import FixAtoms
from agox.candidates.ABC_candidate import disable_cache
import traceback


def relax(model, candidate, optimizer, optimizer_kwargs, optimizer_run_kwargs):
    candidate = candidate.copy()
    candidate.set_calculator(model)
    disable_cache(candidate)
    optimizer = optimizer(candidate,**optimizer_kwargs)
    try:
        optimizer.run(**optimizer_run_kwargs)
    except Exception as e:
        traceback.print_exc()
        print('Relaxation failed with exception: {}'.format(e))
        candidate.add_meta_information('relaxation_steps', -1)
        return candidate
    candidate.add_meta_information('relaxation_steps', 
        optimizer.get_number_of_steps())
    return candidate

class ParallelRelaxPostprocess(PostprocessBaseClass, RayPoolUser):

    """
    Relaxes candidates in parallel using Ray.

    Parameters
    ----------
    model : agox.models.ABC_model
        Model to use for relaxation.
    optimizer : ase.optimize.Optimizer
        Optimizer to use for relaxation.
    optimizer_run_kwargs : dict
        Keyword arguments to pass to the optimizer.run method.
    optimizer_kwargs : dict
        Keyword arguments to pass to the optimizer constructor.
    fix_template : bool
        If True, the template atoms are fixed during relaxation.
    constraints : list
        List of constraints to apply during relaxation.
    start_relax : int
        The iteration to start relaxing candidates.    
    """

    name = 'PoolRelaxer'
    
    def __init__(self, model=None, optimizer=None, optimizer_kwargs={'logfile':None},
        optimizer_run_kwargs={'fmax':0.05, 'steps':200}, fix_template=True,
        constraints=[], start_relax=1, 
        **kwargs):
        RayPoolUser.__init__(self)
        PostprocessBaseClass.__init__(self, **kwargs)

        self.optimizer = BFGS if optimizer is None else optimizer
        self.optimizer_kwargs = optimizer_kwargs
        self.optimizer_run_kwargs = optimizer_run_kwargs
        self.start_relax = start_relax
        self.model = model
        self.constraints = constraints
        self.fix_template = fix_template
        self.model_key = self.pool_add_module(model)

    def process_list(self, candidates):
        """
        Relaxes the given candidates in parallel using Ray. 

        Parameters
        ----------
        list_of_candidates : listn
            List of AGOX candidates to relax. 

        Returns
        -------
        list
            List of relaxed candidates.
        """

        if len(candidates) == 0:
            return candidates

        # Apply constraints. 
        [self.apply_constraints(candidate) for candidate in candidates]

        # Make args, kwargs and modules lists:
        tasks = []
        for candidate in candidates:
            # Create task: function, modules, args, kwargs
            task = Task(function=relax, 
                        modules=[self.model_key], # The model is stored on the actors, tell the task where to find it.
                        args=[candidate, self.optimizer, self.optimizer_kwargs, self.optimizer_run_kwargs], 
                        kwargs={})
            tasks.append(task)

        relaxed_candidates = self.task_map(tasks)
        
        # Remove constraints & move relaxed positions to input candidates:
        # This is due to immutability of the candidates coming from pool_map.
        [self.remove_constraints(candidate) for candidate in candidates]
        for cand, relax_cand in zip(candidates, relaxed_candidates):
            cand.set_positions(relax_cand.positions)
            results = {prop: val for prop, val in relax_cand.calc.results.items() if prop in all_properties}
            cand.calc = SinglePointCalculator(cand, **results)
            cand.meta_information.update(relax_cand.info)

        # Remove if relaxation failed:
        candidates = [cand for cand, rcand in zip(candidates, relaxed_candidates) if rcand.get_meta_information('relaxation_steps') > 0]

        steps = np.array([candidate.get_meta_information('relaxation_steps')
            for candidate in relaxed_candidates])
        self.writer(f'{len(relaxed_candidates)} candidates relaxed.')
        self.writer('   Average number of steps: {:.1f}'.format(steps.mean()))
        self.writer('   Std. number of steps: {:.1f}'.format(steps.std()))


        forces = np.array([candidate.get_forces() for candidate in relaxed_candidates])
        max_forces = np.linalg.norm(forces, axis=-1).max(axis=-1).mean()
        self.writer(f'   Mean max force: {max_forces.mean():.3f}')
        self.writer(f'   Std. max force: {max_forces.std():.3f}')


        return candidates

    def postprocess(self, candidate):
        raise NotImplementedError('"postprocess"-method is not implemented, use postprocess_list.')

    def do_check(self, **kwargs):
        return self.check_iteration_counter(self.start_relax) * self.model.ready_state

    ####################################################################################################################
    # Constraints
    ####################################################################################################################

    def apply_constraints(self, candidate):
        constraints = [] + self.constraints
        if self.fix_template:
            constraints.append(self.get_template_constraint(candidate))

        for constraint in constraints:
            if hasattr(constraint, 'reset'):
                constraint.reset()

        candidate.set_constraint(constraints)

    def remove_constraints(self, candidate):
        candidate.set_constraint([])

    def get_template_constraint(self, candidate):
        return FixAtoms(indices=np.arange(len(candidate.template)))



import os
import pickle
import numpy as np 
from agox.test.test_utils import compare_candidates
from agox.samplers import SamplerBaseClass
from agox.environments import Environment
from ase.io import read, write
from agox.test.test_utils import test_data_dicts, test_folder_path, check_file_is_deleted, get_test_sampler, get_test_environment
from agox.candidates import StandardCandidate

def create_and_compare(generator, sampler, environment, expected_candidates, 
    tolerance, iterations=1, seed=1):
    np.random.seed(seed)

    list_of_list_of_candidates = []
    for _ in range(iterations):
        list_of_list_of_candidates.append(generator(sampler, environment))

    if expected_candidates is not None:
        for list_of_candidates, list_of_expected in zip(list_of_list_of_candidates, expected_candidates):
            for candidate, expected_candidate in zip(list_of_candidates, list_of_expected):
                assert compare_candidates(candidate, expected_candidate, tolerance)

    return list_of_list_of_candidates

def get_test_data(path, environment):
    template = environment.get_template()
    atoms_data = read(path, index=':')
    candidate_data = [StandardCandidate.from_atoms(template, atoms) for atoms in atoms_data]
    return candidate_data

def get_name(generator, dataset_name, parameter_index):
    folder  = 'expected_outputs/'
    name = f'{test_folder_path}generator_tests/{folder}{generator.name}_data{dataset_name}_parameter{parameter_index}.pckl'
    return name

def save_expected_data(data, generator, dataset_name, parameter_index):
    name = get_name(generator, dataset_name, parameter_index)
    check_file_is_deleted(name)
    with open(name, 'wb') as f:
        pickle.dump(data, f)

def load_expected_data(generator, dataset_name, parameter_index):
    name = get_name(generator, dataset_name, parameter_index)
    with open(name, 'rb') as f:
        data = pickle.load(f)

    return data

def generator_testing(generator_class, test_data_dict, generator_args, base_kwargs, 
    update_kwargs, parameter_index, tolerance, seed, test_mode=True):

    # Unpack data dictionary
    path = test_data_dict['path']
    remove = test_data_dict['remove']
    dataset_name = test_data_dict['name']

    # Setup required things:
    environment = get_test_environment(path, remove)
    data = get_test_data(path, environment)
    sampler = get_test_sampler(data)

    # Create the generator
    generator_kwargs = base_kwargs.copy()
    generator_kwargs.update(update_kwargs)
    generator = generator_class(*generator_args, **environment.get_confinement(), **generator_kwargs)

    # Load expected data:
    if test_mode:
        expected_data = load_expected_data(generator, dataset_name, parameter_index)
    else:
        expected_data = None
    
    # Run the actual test. 
    output = create_and_compare(generator, sampler, environment, expected_data, 
        tolerance, iterations=5, seed=seed)

    if not test_mode:
        save_expected_data(output, generator, dataset_name, parameter_index)

    return output






import numpy as np
from agox.writer import agox_writer
from agox.samplers.ABC_sampler import SamplerBaseClass
from agox.observer import Observer
from ase.io import read, write
from copy import deepcopy

class GeneticSampler(SamplerBaseClass):

    """
    Genetic sampler for selecting candidates based on a genetic algorithm.

    Parameters
    ----------
    comparator : callable
        Function that compares two candidates and returns True if they are similar.
    population_size : int
        Number of candidates in the population.
    gets : dict
        Dictionary with the gets that the sampler will observe.
    order : int
        Order of the observer.
    using_add_all : bool
        If False only the final candidates from a trajectory will be considered.
    """

    name = 'GeneticSampler'

    def __init__(self, comparator=None, population_size=10, gets={'get_key':'evaluated_candidates'}, order=6, 
                using_add_all=True, **kwargs):        
        super().__init__(gets=gets, order=order, **kwargs)

        assert comparator is not None
        self.comparator = comparator
        self.population = []
        self.population_size = population_size
        self.sample = self.population
        self.idx = 0
        self.using_add_all = using_add_all

    def get_parents(self, number_of_parents = 1, replace = True):
        if len(self.sample) == 0:
            return []

        P = self.get_selection_probabilities()
        idx = np.random.choice(np.arange(len(self.population)), 1, p=P)[0]
        member = self.population[idx]
        self.increase_times_used(member)
        member = member.copy()
        self.population[idx].copy_calculator_to(member)
        return [member]

    def increase_times_used(self, member):
        count = member.get_meta_information('used_count')
        if count is None:
            count = 0
        else:
            count += 1
        member.add_meta_information('used_count', count)

    def count_similar(self, member, all_candidates):

        count = 0
        for candidate in all_candidates:            
            if candidate.get_meta_information('final'):
                if self.comparator(member, candidate):
                    count += 1
            
        member.add_meta_information('similar_count', count)

    def get_recent_candidates(self, state):
        evaluated_candidates = state.get_from_cache(self, self.get_key)

        if not self.using_add_all:
            return evaluated_candidates
        else:
            recent = []
            for candidate in evaluated_candidates:
                if candidate.get_meta_information('final'): # This is True for the final candidate from a trajectory with LocalOpt evaluator
                    recent.append(deepcopy(candidate))

            E = [atoms.get_potential_energy() for atoms in recent]
            return recent


    @agox_writer
    @Observer.observer_method
    def setup_sampler(self, database, state):
        if self.do_check():
            possible_candidates = self.get_recent_candidates(state)
            all_candidates = database.get_all_candidates()
            self.setup(all_candidates, possible_candidates)
        self.sample = self.population # Lazy renaming, as sample-attribute is used by some methods of the base-class.

    def setup(self, all_candidates, possible_candidates):        
        # If the population is empty we just take all of them for now: 
        if len(self.population) < self.population_size:
            self.population += [possible_candidates.pop(0) for i in range(min(self.population_size, len(possible_candidates)))]
            self.sort_population()
            if len(possible_candidates) == 0:
                return 

        # Now we start deciding whether to replace candidates in the population: 
        for candidate in possible_candidates:
            state = self.consider_candidate(candidate)
            if state:
                self.sort_population()
        
        # Compare population to database:
        for member in self.population:
            self.count_similar(member, all_candidates)

        if self.get_iteration_counter() % 5 == 0:
            write('population_{}.traj'.format(self.get_iteration_counter()), self.population)

        self.print_information()

    def consider_candidate(self, candidate):        
        fitness = self.get_fitness(candidate)

        worst_fitness = self.get_fitness(self.population[-1])

        if fitness < worst_fitness and len(self.population) == self.population_size:
            return False

        for i, member in enumerate(self.population):
            if self.comparator(candidate, member):
                if fitness > self.get_fitness(member):

                    used_count = member.get_meta_information('used_count')
                    
                    del self.population[i]
                    candidate.add_meta_information('used_count', used_count) # Replaced this member so inherit its use count. 
                    self.population.append(candidate)
                    return True
                # If it looks like another member we return regardless of whether it replaces that member. 
                return False

        # If it doesn't look like anything we just replace the worst member of the population:
        del self.population[-1]
        self.population.append(candidate)
    
        return True

    def get_fitness(self, candidate):
        population_energies = [candidate.get_potential_energy() for candidate in self.population]
        e_min = np.min(population_energies)
        e_max = np.max(population_energies)

        p = (candidate.get_potential_energy() - e_min) / (e_max - e_min)

        F = 0.5 * (1 - np.tanh(2*p - 1))

        return F
    
    def sort_population(self):
        if len(self.population):
            fitness = [self.get_fitness(candidate) for candidate in self.population]
            sort_idx = np.argsort(fitness)
            self.population = [self.population[i] for i in sort_idx][::-1]


    def get_selection_probabilities(self):
        N = np.array([member.get_meta_information('used_count') for member in self.population])
        N[N == None] = 0
        N = N.astype(int)
        M = np.array([member.get_meta_information('similar_count') for member in self.population])
        M[M == None] = 0
        M = M.astype(int)
        F = np.array([self.get_fitness(member) for member in self.population])
        U = 1 / np.sqrt(M+1) * 1 / np.sqrt(N + 1)
        P = F * U 
        P = P / np.sum(P)
        return P 

    def print_information(self):
        probs = self.get_selection_probabilities()
        for i, member in enumerate(self.population):
            E = member.get_potential_energy()
            F = self.get_fitness(member)
            P = probs[i]
            self.writer('Member {}: E = {:6.4f}, F = {:6.4f}, P = {:4.2f}'.format(i, E, F, P))

class DistanceComparator:

    def __init__(self, descriptor, threshold):
        self.descriptor = descriptor
        self.threshold = threshold

    def __call__(self, candidate_A, candidate_B):
        return self.compare_candidates(candidate_A, candidate_B)

    def compare_candidates(self, candidate_A, candidate_B):
        feature_A = self.get_feature(candidate_A)
        feature_B = self.get_feature(candidate_B)

        return np.linalg.norm(feature_A-feature_B) < self.threshold

    def get_feature(self, candidate):
        feature = candidate.get_meta_information('population_feature')

        if feature is None:
            feature = self.descriptor.get_features(candidate).sum(axis=0)
            candidate.add_meta_information('population_feature', feature)
        
        return feature





from typing import Tuple

import numpy as np


def plane_to_indices(plane: str) -> Tuple[int, int, int, int]:
    """Convert a string representation of a plane to a 4-tuple of indices
    usable to perform operations using this plane as a projection.

    Parameters
    ----------
    plane : str
        String representation of the plane, in the format /[xyz]{2}[+-]?/. The
        first two characters define the plane's Cartesian axes (and should not
        be equal), and the third optional character defines whether the
        perpendicular axis is positive (+) or negative (-); the default is
        positive.

        Examples of string representations: 'xy', 'xy+', 'xy-', 'xz', 'zx+'.

    Returns
    -------
    Tuple[int, int, int, int]
        A tuple (d0, d1, d2, kv) describing this plane:
        - d0: index of the first plane axis (0, 1, or 2);
        - d1: index of the second plane axis (0, 1, or 2);
        - d2: index of the axis perpendicular to the plane (0, 1, or 2);
        - kv: sign of the axis perpendicular to the plane (-1 or 1).

    Raises
    ------
    ValueError
        Raised if the plane specification is invalid.
    """
    if not isinstance(plane, str):
        plane = str(plane)

    if len(plane) < 2:
        raise ValueError(f'Invalid plane specification {plane}')

    I = np.eye(3)

    d0 = ord(plane[0]) - ord('x')  # 'x' -> 0, 'y' -> 1, 'z' -> 2
    d1 = ord(plane[1]) - ord('x')  # 'x' -> 0, 'y' -> 1, 'z' -> 2

    if len(plane) > 2:
        s = -(ord(plane[2]) - ord(','))  # '+' -> 1, '-' -> -1
    else:
        s = 1

    if d0 not in [0, 1, 2] or d1 not in [0, 1, 2] or d0 == d1 or s not in [-1, 1]:
        raise ValueError(f'Invalid plane specification {plane}')

    d2 = ({0, 1, 2} - {d0} - {d1}).pop()

    i, j = I[d0, :], I[d1, :]
    k = np.cross(i, j)
    kv = s * int(np.sum(k))  # -1 or 1

    return (d0, d1, d2, kv)



import pytest
from agox.models.GPR import SparseGPREnsemble
from agox.models.GPR.kernels import RBF, Constant as C, Noise
from agox.models.descriptors import SOAP
import numpy as np
from agox.test.test_utils import test_data_dicts, get_test_environment, get_test_data

@pytest.mark.ray
@pytest.mark.parametrize('test_data_dict', test_data_dicts)
def test_ray_sgpr_ensemble(test_data_dict, ray_fix, cmd_options):

    # Tolerances:
    tolerance = cmd_options['tolerance']

    path = test_data_dict['path']
    remove = test_data_dict['remove']

    environment = get_test_environment(path, remove)
    data = get_test_data(path, environment)

    train_data = data[0:25]
    test_data = data[25:35]

    # Make a global model: 
    descriptor = SOAP(environment=environment)

    # If the kernel hyperparameters are allowed to take extreme values (large amplitude, small length scale), 
    # it can ocassionally fail - even when the hyperparameters are properly transferred to the actors.
    # But I think that has to be a numerical issue? 
    kernel = C(1, (1, 100)) * RBF(10, (10, 50))

    model = SparseGPREnsemble(descriptor=descriptor,
                kernel=kernel,
                n_optimize=2,
                use_ray=True,
                unc_jitter=1e-3, 
                N_ensemble=10,
                prior_noise=0.1, 
                target_noise=0.05)

    model.train(train_data)
    model.pool.synchronize_module(model)

    # Check properties on actors: 
    result_dicts = model.pool.get_module_attributes(model, model.dynamic_attributes)

    for result_dict in result_dicts:
        for attribute_name in ['X', 'Xm', "alpha", "ensemble_alpha", "priors_per_atom", "mean_energy"]:
            np.testing.assert_allclose(result_dict[attribute_name], model.__getattribute__(attribute_name), **tolerance)

        # Kernel check seperately because it is annoying. 
        np.testing.assert_allclose(result_dict['kernel'].theta, model.kernel.theta, **tolerance)

    # Predictions without Ray: 
    energies = np.array(model.predict_energy(test_data))
    forces = np.array(model.predict_forces(test_data))
    uncertainties = np.array(model.predict_uncertainty(test_data))
    unceertainty_forces = np.array(model.predict_uncertainty_forces(test_data))

    # Predictions with Ray:
    def predict(model, atoms): 
        E = np.array(model.predict_energy(atoms))
        F = np.array(model.predict_forces(atoms))
        sigma = np.array(model.predict_uncertainty(atoms))
        sigma_F = np.array(model.predict_uncertainty_forces(atoms))

        results = {'E': E, 'F': F, 'sigma': sigma, 'sigma_F': sigma_F}
        return results
    pool = model.pool
    # Make predictions on all actors:
    key = pool.get_key(model)
    results = pool.execute_on_actors(predict, [key], [test_data], {})

    # Check that all predictions on every actor are the same:
    results_actor_0 = results[0]
    for key in results[0].keys():
        for results_actor in results[1:]:
            np.testing.assert_allclose(results_actor_0[key], results_actor[key])

    # # Check that predictions are the same as without Ray:
    np.testing.assert_allclose(energies, results_actor_0['E'], **tolerance)
    np.testing.assert_allclose(forces, results_actor_0['F'], **tolerance)
    np.testing.assert_allclose(uncertainties, results_actor_0['sigma'], **tolerance)
    np.testing.assert_allclose(unceertainty_forces, results_actor_0['sigma_F'], **tolerance)



import numpy as np
from copy import copy
from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass
from ase.data import covalent_radii

class Voronoi(DescriptorBaseClass):

    name = "Voronoi"
    descriptor_type = "global"

    """
    Voronoi graph descriptor.

    Parameters
    ----------
    indices : list
        List of indices of atoms to be considered.
        If None and template is None, all atoms are considered.
        If None and template is not None, all atoms except for template are considered.
    template : Atoms
        Template structure. If None, the first structure in the list of structures is used.
    covalent_bond_scale_factor : float  
        Scaling of covalent bond length to determine the cutoff radii.
    n_points : int
        Number of points sampled from circle on plane.
    angle_from_central_atom : float
        Angle from central atom to determine the points sampled from circle on plane.
    """

    def __init__(
        self,
        indices=None,
        template=None,
        covalent_bond_scale_factor=1.3,
        n_points = 8,
        angle_from_central_atom = 20,
        *args,
        **kwargs):

        super().__init__(*args, **kwargs)

        self.indices = indices
        self.template = template
        self.covalent_bond_scale_factor = covalent_bond_scale_factor
        self.n_points = n_points
        self.angle_from_central_atom = angle_from_central_atom

    def get_bond_matrix(self,candidate) -> np.ndarray:
        """
        Calculate the bond matrix for a candidate structure.

        Parameters
        ----------
        candidate : ase.Atoms
            Atoms object.
        
        Returns
        -------
        bond matrix: np.ndarray
            Array of shape [n_atoms, n_atoms]
        """
        candidate_copy = candidate.copy()
        candidate_copy.set_constraint()

        del candidate_copy[[i for i in range(len(candidate)) if i not in self.indices]]

        distance_vectors = candidate_copy.get_all_distances(mic = True, vector = True)
        distance_abs = np.linalg.norm(distance_vectors, axis = 2)
        numbers = candidate_copy.get_atomic_numbers()
        r = [covalent_radii[number] for number in numbers]
        x,y = np.meshgrid(r,r)
        optimal_distances = x+y
    
        distances_rel = distance_abs / optimal_distances        

        matrix = np.logical_and(distances_rel > 1e-3, distances_rel < self.covalent_bond_scale_factor).astype(int)
        matrix += np.diag(numbers)

        for i, (distances, vectors, position) in enumerate(zip(distance_abs, distance_vectors, candidate_copy.positions)):
            planes = []

            args = np.argsort(distances)[1:]
            distances = np.expand_dims(distances, 1)
            distances = distances[args]
            normal_vectors = vectors[args] / distances 
            points = 0.5 * distances * normal_vectors + position
            for j, normal_vector, point, distance in zip(args, normal_vectors, points, distances):
                if i != j and matrix[i, j] == 1:   
                    if len(planes) == 0:
                        plane = Plane(normal_vector, point)
                        planes.append(plane)
                    else:
                        cuts = [plane.plane_cut(point) for plane in planes]
                        cuts = np.any(cuts)

                        plane = Plane(normal_vector, point)

                        if cuts:
                            matrix[i, j] = 0
                        elif self.n_points != 0 and self.angle_from_central_atom != 0:
                            points_in_plane = plane.points_in_plane(self.n_points, 0.5 * distance, self.angle_from_central_atom)
                            for p in points_in_plane:
                                cuts = [plane.plane_cut(p) for plane in planes]
                                cuts = np.any(cuts)
                                if cuts:
                                    matrix[i, j] = 0
                                    break

                        planes.append(plane)                       

        matrix = np.floor((matrix + matrix.T)/2)

        return matrix

    def convert_matrix_to_eigen_value_string(self, matrix) -> str:
        '''
        Convert the bond matrix to a string of sorted eigenvalues.

        Parameters
        ----------
        matrix : np.ndarray
            Bond matrix.
        
        Returns
        -------
        eigenvalue string : str
            String of sorted eigenvalues.
        '''
        w,_ = np.linalg.eig(matrix)
        w = np.real(w)
        w.sort()
        s = '[' + ','.join(['{:8.3f}'.format(e) for e in w]) + ']'
        s = s.replace('-0.000',' 0.000')
        return s

    def create_features(self, t, allow_read_from_meta_information=False) -> str:
        '''
        Calculate the Voronoi graph descriptor.

        Parameters
        ----------
        t : ase.Atoms
            Atoms object.
        allow_read_from_meta_information : bool
            Whether to allow reading the eigenvalue string from the meta information of the Atoms object.

        Returns
        -------
        eigenvalue string : str
            String of sorted eigenvalues.
        '''

        # this is conditioned that the meta information is trusted (corrupted when Candidates are copied and changed)
        if allow_read_from_meta_information:
            if t.has_meta_information('eigen_string'):
                return t.get_meta_information('eigen_string')

        if self.indices is None and self.template is not None:
            self.indices = list(range(len(self.template), len(t)))
            print('SETTING INDICES:',self.indices)
        elif self.indices is None and self.template is None:
            self.indices = list(range(len(t)))

        m = self.get_bond_matrix(t)

        s = self.convert_matrix_to_eigen_value_string(m)
        try:
            t.add_meta_information('eigen_string',s)
        except:
            pass

        return s

    def get_number_of_centers(self, atoms):
        """
        Get the number of eigenvalues for Voronoi graph descriptor.

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms object.
        
        Returns
        -------
        n_centers : int
            Number of eigenvalues.    
        """
        return len(atoms)

class Plane():
    '''
    Class for constructing planes to determine whether or not two atoms
    are bound to each other.

    Parameters:
    ------------
    nvec: Numpy array
        Normal vector of the plane
    p0: Numpy array
        Point in the plane
    '''

    def __init__(self, nvec, p0):
        self.nvec = nvec 
        self.p0 = p0 

    def plane_cut(self, p) -> bool:
        '''
        Determine whether a point is cut by the plane or not.

        Parameters
        ----------
        p : Numpy array
            Point to be checked.

        Returns
        -------
        cut : bool
            Whether the point is cut by the plane (True) or not (False).
        '''
        return np.sum(self.nvec * (p - self.p0)) > 0
    
    def points_in_plane(self, n_points, distance_from_atom_to_p0 = 1, angle_from_central_atom = 10) -> list:
        '''
        Generate points on the circle on the plane.

        Parameters
        ----------
        n_points : int
            Number of points to be generated.
        distance_from_atom_to_p0 : float
            Distance from atom to the point p0.
        angle_from_central_atom : float
            Angle from central atom to determine the points sampled from circle on plane.

        Returns
        -------
        points : list
            List of points on the circle on the plane.
        '''

        index1 = np.argmax(np.abs(self.nvec))
        if index1 != 0:
            index2= 0
        else:
            index2 = 1

        p1 = np.zeros(3)
        p1[index2] = 1
        p1[index1] = - self.nvec[index2] / self.nvec[index1]

        p1 /= np.linalg.norm(p1)
        p2 = np.cross(self.nvec, p1)

        angle = 2 * np.pi / n_points
        
        points = [] 
        for i in range(n_points):
            p = np.cos(angle * i) * p1 + np.sin(angle * i) * p2
            p *= distance_from_atom_to_p0 * np.tan(np.deg2rad(angle_from_central_atom))
            p += self.p0
            points.append(p)

        return points



import warnings
from typing import Dict, List, Optional, Tuple

import numpy as np
from ase import Atoms
from scipy.linalg import lstsq, qr
from scipy.sparse import dia_matrix

from agox.models.GPR.GPR import GPR
from agox.utils import candidate_list_comprehension
from agox.utils.filters import AllFilter, NoneFilter
from agox.utils.sparsifiers import CUR, SparsifierBaseClass

from timeit import default_timer as dt


class SparseGPR(GPR):

    """
    Sparse Gaussian Process Regression with inducing points.

    Parameters:
    -----------
    sigma : float
        Noise level for both energy and forces data, by default 0.01
    sigma_E : float
        Noise level for energy data is per descriptor center ie. per atom for local
        descriptors and per configuration for global descriptors.
    sigma_F : float
        Noise level for force data.
    sparsifier : SparsifierBaseClass
        Sparsifier object
    force_data_filter : FilterBaseClass
        Filter object for force data. Selection of atoms from which the forces are
        used for training.
    jitter : float
        Jitter level
    train_uncertainty : bool
        If True the matrices needed to predict uncertainty are trained, by default False

    noise : float
        Deprecated. Use sigma instead
    noise_E : float
        Deprecated. Use sigma_E instead
    noise_F : float
        Deprecated. Use sigma_F instead
        
    """


    name = "SparseGPR"

    supported_descriptor_types = ["global", "local"]

    implemented_properties = ["energy", "forces", "local_energy", "uncertainty"]

    def __init__(
        self,
        sigma: float = 0.01,
        centralize: bool = False,
        jitter: float = 1e-8,
        unc_jitter: float = 1e-5,
        filter=None,
        sparsifier: SparsifierBaseClass = CUR(),
        n_optimize=0,
        force_data_filter=NoneFilter(),
        noise: Optional[float] = None,
        sigma_E: Optional[float] = None,
        sigma_F: Optional[float] = None,
        noise_E: Optional[float] = None,
        noise_F: Optional[float] = None,
        sparsification_schedule = None,
        train_uncertainty: bool = False,
        **kwargs
    ) -> None:
        """

        Parameters
        ----------
        """
        super().__init__(
            centralize=centralize, filter=filter, n_optimize=n_optimize, **kwargs
        )

        for attr in ["X", "Xm", "alpha", "kernel", "mean_energy", "_ready_state"]:
            self.add_dynamic_attribute(attr)

        self.jitter = jitter
        self.unc_jitter = unc_jitter

        # Deprecation warnings:
        if noise is not None:
            warnings.warn(
                "noise is deprecated and will be removed in version 4.0.0. Use sigma instead", DeprecationWarning
            )
            sigma = noise
        if noise_E is not None:
            warnings.warn(
                "noise_E is deprecated and will be removed in version 4.0.0. Use sigma_E instead", DeprecationWarning
            )
            sigma_E = noise_E
        if noise_F is not None:
            warnings.warn(
                "noise_F is deprecated and will be removed in version 4.0.0. Use sigma_F instead", DeprecationWarning
            )
            sigma_F = noise_F

        if sigma_E is None:
            self.sigma_E = sigma
        else:
            self.sigma_E = sigma_E

        if noise_F is None:
            self.sigma_F = sigma
        else:
            self.sigma_F = sigma_F

        self.sparsifier = sparsifier

        self._transfer_data = []
        self._transfer_noise = np.array([])

        if force_data_filter == "all":
            self.force_data_filter = AllFilter()
        elif force_data_filter == "none":
            self.force_data_filter = NoneFilter()
        else:
            self.force_data_filter = force_data_filter

        self.train_uncertainty = train_uncertainty

        if self.train_uncertainty:
            self.add_dynamic_attribute("C_inv")
            self.add_dynamic_attribute("Kmm_inv")

        if self.use_ray:
            self.actor_model_key = self.pool_add_module(self)
            self.self_synchronizing = True  # Defaults to False, inherited from Module.

        if sparsification_schedule is not None:
            self.sparsification_schedule = sparsification_schedule

        self._set_saved_attributes()

    def _set_saved_attributes(self):
        self.reset_save_attributes()
        self.add_save_attributes(['Xm', 'X', 'alpha', 'single_atom_energies', '_ready_state', 'kernel.theta', 'mean_energy'])
        if self.train_uncertainty:
            self.add_save_attributes(['C_inv', 'Kmm_inv'])

    @candidate_list_comprehension
    def predict_local_energy(self, atoms: Atoms, **kwargs) -> np.ndarray:
        """
        Calculate the local energies in the model.

        Parameters
        ----------
        atoms : ase.Atoms
            ase.Atoms object
        X : np.ndarray
            Features for the ase.Atoms object

        Returns
        -------
        np.ndarray
            Local energies

        """
        X = self._get_features(atoms)
        k = self.kernel(self.Xm, X)
        return (k.T @ self.alpha).reshape(
            -1,
        ) + self.single_atom_energies[atoms.get_atomic_numbers()]

    @candidate_list_comprehension
    def predict_uncertainty(
        self, atoms: Atoms, k: np.ndarray = None, k0: np.ndarray = None, **kwargs
    ) -> float:
        """
        Predict uncertainty for a given atoms object

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms object

        Returns
        -------
        float
            Uncertainty

        """
        if "uncertainty" not in self.implemented_properties or not self.ready_state:
            unc = 0
        else:
            cov = self.predict_covariance(atoms, k=k, k0=k0)
            unc = np.sqrt(cov.sum())
        return unc

    def predict_variances(self, atoms: Atoms, k: np.ndarray = None, k0: np.ndarray = None, **kwargs):
        cov = self.predict_covariance(atoms, k=k, k0=k0)
        local_variances = np.diag(cov)
        return local_variances
    
    @candidate_list_comprehension
    def predict_covariance(self, atoms: Atoms, k: np.ndarray = None, k0: np.ndarray = None, **kwargs):
        """
        Calculate the covariance matrix of local energies for a given atoms object.
        """
        X = self._get_features(atoms)

        if k is None:
            k = self.kernel(self.Xm, X)

        if k0 is None:
            k0 = self.kernel(X, X)

        cov = k0 - k.T @ self.Kmm_inv @ k + k.T @ self.C_inv @ k

        return cov

    @candidate_list_comprehension
    def predict_uncertainty_forces(self, atoms, x=None, k=None, dk_dr=None, dx_dr=None, **kwargs):

        if "uncertainty" not in self.implemented_properties or not self.ready_state:
            return np.zeros((len(atoms), 3))
        
        # Compute the required things if they are not supplied: 
        if k is None:
            if x is None:
                x = self._get_features(atoms)
            k = self.kernel(self.Xm, x)

        if dx_dr is None:
            dx_dr = self.descriptor.get_feature_gradient(atoms)

        if dk_dr is None:
            dk_dr = self._get_kernel_derivative(atoms, x=x, dX_dr=dx_dr)

        dk0_dx = np.array([self.kernel.get_feature_gradient(x, x[i : i + 1]) for i in range(len(x))])

        # Kmm_inv % C_inv part.
        # This is likely the optimal for all realistic cases (unless number of atoms and number of m-points is close)
        # Small optimization my moving subtraction of the to m x m matrices to the 
        # training function, but ehh. 
        opt_ein_sum_path = ['einsum_path', (1, 2), (0, 1)]
        forces = 2 * np.einsum('ijdm, mn, no -> jd', dk_dr, (self.C_inv - self.Kmm_inv), k, optimize=opt_ein_sum_path)
        
        # k0-part:
        forces += np.einsum('indf,ijf->nd', dx_dr, dk0_dx) - np.einsum('jndf,ijf->nd', dx_dr, dk0_dx)

        # Total std is required too:
        std = self.predict_uncertainty(atoms, x=x, k=k)

        return -forces / (2 * std)

    @property
    def noise_E(self) -> float:
        """
        Noise level

        Returns
        -------
        float
            Noise level

        """
        return self._noise

    @noise_E.setter
    def noise_E(self, s: float) -> None:
        """
        Noise level

        Parameters
        ----------
        s : float
            Noise level

        """
        self._noise = s

    @property
    def transfer_data(self) -> List[Atoms]:
        """
        List of ase.Atoms objects to transfer to the model

        Returns
        -------
        list of ase.Atoms
            List of ase.Atoms objects to transfer to the model

        """
        return self._transfer_data

    @transfer_data.setter
    def transfer_data(self, l: List[Atoms]) -> None:
        """
        List of ase.Atoms objects to transfer to the model

        Parameters
        ----------
        l : list of ase.Atoms
            ase.Atoms objects to transfer to the model

        """
        warnings.warn(
            "transfer_data should not be used. Use add_transfer_data instead."
        )
        self.add_transfer_data(l)

    @property
    def transfer_noise(self) -> np.ndarray:
        """
        Noise level for transfer data

        Returns
        -------
        np.ndarray
            Noise level for transfer data

        """
        return self._transfer_noise

    @transfer_noise.setter
    def transfer_noise(self, s: np.ndarray) -> None:
        """
        Noise level for transfer data

        Parameters
        ----------
        s : np.ndarray
            Noise level for transfer data

        """
        warnings.warn(
            "transfer_noise should not be used. Use add_transfer_data instead."
        )
        self._transfer_noise = s

    def add_transfer_data(self, data: List[Atoms], noise: float = None) -> None:
        """
        Add ase.Atoms objects to the transfer data

        Parameters
        ----------
        data : list of ase.Atoms
            List of ase.Atoms objects to add to the transfer data
        noise : float
            Noise level for the transfer data

        """
        if isinstance(data, list):
            self._transfer_data += data
            if noise is None:
                noise = self.sigma_E

            self._transfer_noise = np.append(
                self._transfer_noise, np.ones(len(data)) * noise
            )

        else:
            self.add_transfer_data.append([data])

    def get_all_data(self, data: List[Atoms]) -> List[Atoms]:
        """
        Prepend transfer data to the data for use when training. 
        """
        return self.transfer_data + data

    def model_info(self, **kwargs) -> List[str]:
        """
        List of strings with model information
        """
        x = "    "

        filter_name = self.filter.name if self.filter is not None else "None"
        try:
            data_before_filter = self._data_before_filter
            data_after_filter = self._data_after_filter
            filter_removed_data = self._data_before_filter - self._data_after_filter
        except AttributeError:
            filter_removed_data = 0

        sparsifier_name = (
            self.sparsifier.name if self.sparsifier is not None else "None"
        )
        sparsifier_mpoints = (
            self.sparsifier.m_points if self.sparsifier is not None else "None"
        )
        force_data_filter = (
            self.force_data_filter.name if self.force_data_filter else "None"
        )

        out = [
            "------ Model Info ------",
            "Descriptor:",
            x + "{}".format(self.descriptor.name),
            "Kernel:",
            x + "{}".format(self.kernel),
            "Filter:",
            x + "{} removed {} structures".format(filter_name, filter_removed_data),
            x + "- Data before filter: {}".format(data_before_filter),
            x + "- Data after filter: {}".format(data_after_filter),
            "Sparsifier:",
            x
            + "{} selecting {} inducing points".format(
                sparsifier_name, sparsifier_mpoints
            ),
            "Force-data filter:",
            x + "{}".format(force_data_filter),
            "Noise:",
            x + "Energy noise: {}".format(self.sigma_E),
            x + "Forces noise: {}".format(self.sigma_F),
            "------ Training Info ------",
            "Total training data: {} energies and {} forces".format(
                self._N_energies, self._N_forces
            ),
            "Transfer data size: {}".format(len(self.transfer_data)),
            "Number of local environments: {}".format(self.Xn.shape[0]),
            "Number of inducing points: {}".format(self.Xm.shape[0]),
            "Neg. log marginal likelihood.: {:.2f}".format(self._nlml),
        ]

        return out

    def _train_model(self, data: List[Atoms], synchronize=True,
                     train_projected_process=True) -> None:
        """
        Train the model

        """
        assert self.Xn is not None, "self.Xn must be set prior to call"
        assert self.L is not None, "self.L must be set prior to call"
        assert self.Y is not None, "self.Y must be set prior to call"

        t_hyper = dt()
        if self.use_ray:
            self.pool_synchronize(
                attributes=["Xn", "Xm", "Y", "L", "sigma_inv", "Xf", "dXf_dr"],
                writer=self.writer,
            )
            self.hyperparameter_search_parallel(relax=False)
        else:
            self.monte_carlo_hyperparameter_search()
        self.training_timings['hyperparameter_search'] = dt() - t_hyper


        t_train = dt()
        self.K_mm = self.kernel(self.Xm)
        self.K_nm = self.kernel(self.Xn, self.Xm)
       
        all_data = self.get_all_data(data)

        if len(self.training_force_data_indices) > 0:
            training_force_data = [
                all_data[i] for i in self.training_force_data_indices
            ]
            K_nm_jac = np.concatenate(
                [
                    self._get_sparse_kernel_derivative(x=x, dX_dr=dX_dr)
                    for x, dX_dr in zip(self.Xf, self.dXf_dr)
                ]
            )
            self.K_nm = np.concatenate([self.K_nm, K_nm_jac], axis=0)
        else:
            training_force_data = []

        LK_nm = self.L @ self.K_nm
        K = (
            self.K_mm
            + LK_nm.T @ self.sigma_inv @ LK_nm
            + self.jitter * np.eye(self.K_mm.shape[0])
        )
        K = self._symmetrize(K)

        self.alpha = self._solve(K, LK_nm.T @ self.sigma_inv @ self.Y)

        self.training_timings['training'] = dt() - t_train

        train_uncertainty = self.train_uncertainty and train_projected_process
        if train_uncertainty:
            t_uncertainty = dt()
            self.Kmm_inv = self._solve(
                self.K_mm + self.unc_jitter * np.eye(self.K_mm.shape[0])
            )

            local_sigma_inv = self._make_local_sigma(all_data, training_force_data)

            C = (
                self.K_mm
                + self.K_nm.T @ local_sigma_inv @ self.K_nm
                + self.unc_jitter * np.eye(self.K_mm.shape[0])
            )
            C = self._symmetrize(C)
            self.C_inv = self._solve(C)

            self.training_timings['uncertainty'] = dt() - t_uncertainty

        self.ready_state = True
        if self.use_ray and synchronize:
            t_sync = dt()
            attributes_to_sync = ["X", "alpha", "kernel", "mean_energy", "_ready_state"]
            
            if train_uncertainty:
                attributes_to_sync += ["Kmm_inv", "C_inv"]
            
            self.pool_synchronize(
                attributes=attributes_to_sync,
                writer=self.writer,
            )

            self.training_timings['synchronize'] = dt() - t_sync

    def _get_sparse_kernel_derivative(
        self, x: np.ndarray, dX_dr: np.ndarray
    ) -> np.ndarray:
        """
        Get the kernel derivative wrt. the Carteisian coordinates r.

        Parameters
        ----------
        x : np.ndarray, optional
            Features, by default None
        dX_dr : np.ndarray, optional
            Derivatives of the features wrt. the Carteisian coordinates r, by default None

        Returns
        -------
        np.ndarray
            Derivatives of the kernel wrt. the Carteisian coordinates r.
        """
        # Get kernel derivatives:
        dk_dX = np.array(
            [
                self.kernel.get_feature_gradient(self.Xm, x[i : i + 1])
                for i in range(len(x))
            ]
        )

        # Convert to derivatives wrt. Carteisian coordinates r.
        dk_dr = np.einsum(
            "nadf,nmf->nadm", dX_dr, dk_dX
        )  # [centers, atoms, 3, features]

        c, a, t, M = dk_dr.shape
        return dk_dr.sum(axis=0).reshape(a * t, M)

    def _get_forces_features(
        self, atoms: List[Atoms]
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """
        Get the features and derivatives of the features for the force data.

        Parameters
        ----------
        atoms : List[ase.Atoms]
            List of Atoms objects.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Features and derivatives of the features.
        """
        x = [self._get_features(a) for a in atoms]
        dX_dr = [self.descriptor.get_feature_gradient(a) for a in atoms]
        return x, dX_dr

    def _log_marginal_likelihood(self, theta: Optional[np.ndarray] = None) -> float:
        """
        Marginal log likelihood

        Parameters
        ----------
        theta : np.ndarray
            Kernel parameters

        Returns
        -------
        float
            log Marginal likelihood

        """
        if theta is not None:
            t = self.kernel.theta.copy()
            self.kernel.theta = theta
            K_nm = self.kernel(self.Xn, self.Xm)

            if len(self.Xf) > 0:
                K_nm_jac = np.concatenate(
                    [
                        self._get_sparse_kernel_derivative(x=x, dX_dr=dX_dr)
                        for x, dX_dr in zip(self.Xf, self.dXf_dr)
                    ]
                )
                K_nm = np.concatenate([K_nm, K_nm_jac], axis=0)

            Kmm_inv = self._solve(
                self.kernel(self.Xm) + self.unc_jitter * np.eye(self.Xm.shape[0])
            )
            self.kernel.theta = t
        else:
            K_nm = self.K_nm
            Kmm_inv = self.Kmm_inv

        LK_nm = self.L @ K_nm
        Ktilde = LK_nm @ Kmm_inv @ LK_nm.T + np.diag(1 / np.diagonal(self.sigma_inv))

        sign, logdet = np.linalg.slogdet(Ktilde)
        if sign <= 0:
            return np.inf

        lml = -0.5 * logdet
        a = self._solve(Ktilde, self.Y[:, 0])
        lml -= 0.5 * self.Y[:, 0].T @ a
        lml -= 0.5 * self.Y.shape[0] * np.log(2 * np.pi)

        return float(lml)

    def _preprocess(self, data: List[Atoms]) -> None:
        """
        Preprocess the training data for the model

        Parameters
        ----------
        data : list of ase.Atoms
            List of ase.Atoms objects

        Returns
        -------
        np.ndarray
            Features for the ase.Atoms objects
        np.ndarray
            Energies for the ase.Atoms objects

        """
        all_data = self.get_all_data(data)
        X, Y = super()._preprocess(all_data)
        self.Xn = X

        training_force_data, self.training_force_data_indices = self.force_data_filter(
            all_data
        )
        self.Xf, self.dXf_dr = self._get_forces_features(training_force_data)

        if len(training_force_data) > 0:
            F = np.concatenate(
                [atoms.get_forces().reshape(-1) for atoms in training_force_data]
            ).reshape(-1, 1)
            self._N_forces = F.shape[0]

            if self.prior is not None:
                F -= np.concatenate(
                    [
                        self.prior.predict_forces(atoms).reshape(-1)
                        for atoms in training_force_data
                    ]
                ).reshape(-1, 1)

            Y = np.concatenate([Y, F])
        else:
            self._N_forces = 0

        self.L = self._make_L(all_data, training_force_data, X.shape)
        self.sigma_inv = self._make_sigma(all_data, training_force_data)

        t_sparsify = dt()
        if self.sparsifier is not None:

            if hasattr(self, 'num_enviornments_when_sparsified'):
                new_environments = self.Xn.shape[0] - self.num_enviornments_when_sparsified

            if not hasattr(self, 'Xm'):
                # First time training we have to do sparsification.
                self.Xm, _ = self.sparsifier(self.Xn)
                self.num_enviornments_when_sparsified = len(self.Xn)
            elif self.get_iteration_counter() is None:
                # If the iteration counter is None, we have to do sparsification,
                # as it is not part of an AGOX run. 
                self.Xm, _ = self.sparsifier(self.Xn)
                self.num_enviornments_when_sparsified = len(self.Xn)

            elif self.sparsification_schedule(total_environments=self.Xn.shape[0],
                                              sparse_environments=self.sparsifier.m_points,
                                              iteration=self.get_iteration_counter(),
                                              new_environments=new_environments):
                self.Xm, _ = self.sparsifier(self.Xn)
                self.num_enviornments_when_sparsified = len(self.Xn)            
        else:
            self.Xm = self.Xn

        self.training_timings['sparsification'] = dt() - t_sparsify

        return self.Xm, Y

    def _make_L(
        self,
        atoms_list: List[Atoms],
        training_force_data: List[Atoms],
        shape_X: Tuple[int, int],
    ) -> np.ndarray:
        """
        Make the L matrix

        Parameters
        ----------
        atoms_list : list of ase.Atoms
            List of ase.Atoms objects

        Returns
        -------
        np.ndarray
            L matrix

        """
        lengths = [self.descriptor.get_number_of_centers(atoms) for atoms in atoms_list]
        r = len(lengths)
        c = np.sum(lengths)

        col = 0
        L = np.zeros((r, c))
        for i, atoms in enumerate(atoms_list):
            L[i, col : col + self.descriptor.get_number_of_centers(atoms)] = 1.0
            col += self.descriptor.get_number_of_centers(atoms)

        if len(training_force_data) > 0:
            cf = np.sum([len(atoms) for atoms in training_force_data])

            E_pad = np.zeros((r, 3 * cf))
            F_pad = np.zeros((3 * cf, c))
            L_F = np.zeros((3 * cf, 3 * cf))
            col = 0
            for i, atoms in enumerate(training_force_data):
                L_F[col : col + 3 * len(atoms), col : col + 3 * len(atoms)] = np.ones(
                    3 * len(atoms)
                )
                col += 3 * len(atoms)

            L_F[np.diag_indices_from(L_F)] = 0.0
            L = np.block([[L, E_pad], [F_pad, L_F]])

        return L

    def _make_sigma(
        self, atoms_list: List[Atoms], training_force_data: List[Atoms]
    ) -> np.ndarray:
        """
        Make the sigma matrix

        Parameters
        ----------
        atoms_list : list of ase.Atoms
            List of ase.Atoms objects

        Returns
        -------
        np.ndarray
            Sigma matrix

        """
        sigmas = np.array(
            [
                self.sigma_E**2 * self.descriptor.get_number_of_centers(atoms)
                for atoms in atoms_list
            ]
        )
        sigmas[: len(self.transfer_data)] = self.transfer_noise**2 * np.array(
            [
                self.descriptor.get_number_of_centers(atoms)
                for atoms in self.transfer_data
            ]
        )
        if len(training_force_data) > 0:
            sigmas_F = self.sigma_F**2 * np.ones(
                (3 * np.sum([len(atoms) for atoms in training_force_data]),)
            )
            sigmas = np.concatenate([sigmas, sigmas_F])

        sigma_inv = np.diag(1 / sigmas)

        return sigma_inv

    def _make_local_sigma(
        self, atoms_list: List[Atoms], training_force_data: List[Atoms]
    ) -> np.ndarray:
        """
        Make the local sigma matrix. This is the inverse of the local noise
        variance.

        Parameters
        ----------
        atoms_list : list of ase.Atoms

        Returns
        -------
        np.ndarray
            Local sigma matrix
        """
        local_sigmas = self.sigma_E**2 * np.ones(self.K_nm.shape[0])
        if len(self.transfer_data) > 0:
            idx = np.sum([len(atoms) for atoms in self.transfer_data])
            n_repeats = [
                self.descriptor.get_number_of_centers(atoms)
                for atoms in self.transfer_data
            ]
            local_sigmas[:idx] = np.repeat(
                np.array(self.transfer_noise) ** 2, n_repeats
            )

        if len(training_force_data) > 0:
            sigmas_F = self.sigma_F**2 * np.ones(
                (3 * np.sum([len(atoms) for atoms in training_force_data]),)
            )
            local_sigmas[-len(sigmas_F) :] = sigmas_F

        # local_sigma_inv = np.diag(1 / local_sigmas)
        local_sigma_inv = dia_matrix((1 / local_sigmas, 0), shape=(len(local_sigmas), len(local_sigmas)))
        return local_sigma_inv

    def _solve(self, A: np.ndarray, Y: Optional[np.ndarray] = None) -> np.ndarray:
        """
        Solve the linear system using QR decomposition and least squares

        Parameters
        ----------
        A : np.ndarray
            Matrix A
        Y : np.ndarray
            Matrix Y

        Returns
        -------
        np.ndarray
            Solution X

        """
        Q, R = qr(A)
        if Y is None:
            return lstsq(R, Q.T)[0]
        else:
            return lstsq(R, Q.T @ Y)[0]

    def _symmetrize(self, A: np.ndarray) -> np.ndarray:
        """
        Symmetrize a matrix

        Parameters
        ----------
        A : np.ndarray
            Matrix to symmetrize

        Returns
        -------
        np.ndarray
            Symmetrized matrix

        """
        return (A + A.T) / 2

    @staticmethod
    def sparsification_schedule(total_environments,
                                sparse_environments,
                                iteration,
                                new_environments):
        """
        Schedule for sparsification. 
        """
        if total_environments < sparse_environments * 10:
            return True
        elif iteration % 25 == 0:
            return True

        return False


from tokenize import Single
from agox.evaluators.single_point import SinglePointEvaluator
import numpy as np


class RattleEvaluator(SinglePointEvaluator):

    def __init__(self, calculator, seed=0, constraints=[], stdev=0.1, skip_frequency=10, **kwargs): 
        super().__init__(calculator=calculator, **kwargs)
        self.seed = seed
        self.constraints = constraints
        self.stdev = stdev
        self.skip_frequency = skip_frequency


    def evaluate_candidate(self, candidate):
        if self.get_iteration_counter() % self.skip_frequency > 0:
            candidate.set_constraint(self.constraints)

            self.writer('Im rattling with stdev:', self.stdev)
                
            seed = self.seed*10000 + self.get_iteration_counter()
            candidate.rattle(stdev=self.stdev, seed=seed)
            candidate.set_constraint()
            
        return super().evaluate_candidate(candidate)



import os

# Versioning
__version_info__ = (3, 8, 0)
__version__ = '{}.{}.{}'.format(*__version_info__)

# Extra versioning - Mainly for managing Pypi releases.
version_extra = os.environ.get('AGOX_VERSION_EXTRA', None)
if version_extra:
    __version__ = '{}{}'.format(__version__, version_extra)

try: # When installing the package we don't need to actually import.
    from agox.module import Module
    from agox.observer import Observer
    from agox.writer import Writer
    from agox.tracker import Tracker
    from agox.main import AGOX

    __all__ = ['Module', 'Observer', 'Writer', 'Tracker', 'AGOX', '__version__']

except ImportError:
    pass






import numpy as np
from ase.calculators.calculator import Calculator
import os
from ase.io import read, write, Trajectory
import importlib.util
import pickle

"""
If the GPAW Calculator throws unexpected errors it is likely because Ray has somehow 
been imported on its process, look at the imports that are happening for it and 
make sure Ray is not present in any file that is imported (even indirectly through 
__init__.py files).
"""

calc_script = """
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
from agox.utils.constraints.box_constraint import BoxConstraint
from agox.main import AGOX

from ase.io import read, write, Trajectory
from gpaw import GPAW, PW
import sys
import argparse
import pickle

# Input stuff:
parser = argparse.ArgumentParser()
parser.add_argument('input_file')
parser.add_argument('output_file')
parser.add_argument('settings_file')

# Read the input file: 
args = parser.parse_args()

t = read(args.input_file)

# Setup the GPAW calculator:
with open(args.settings_file, 'rb') as f:
    settings = pickle.load(f)

#if 'modules' in settings.keys():
modules = settings.pop('modules')
for module in modules:
    exec(module)

# Built settings: 
built_settings = dict()
for key, setting in settings.items():
    try: # Some settings are functions.
        built_settings[key] = eval(setting)
    except: # Some are just strings that wont eval
        built_settings[key] = setting

calc = GPAW(**built_settings)

# Write the output file:
t.set_calculator(calc)
t.get_forces(apply_constraint=False)
t.get_potential_energy(apply_constraint=False)
Trajectory(args.output_file, mode='a').write(t)
"""

test_calc_script="""
print('I HAVE BEEN SUMMONED')
"""

class GPAW_IO(Calculator):
    """
    GPAW IO Calculator

    parexe: bool
        Run in parallel or not.
    script_name: str
        Name of the script file that will be written and execute. 
    settings_name: str
        Name of pickle file that settings will be written to. 
        If more than one calculator is used at the same time, each should have a distinct settings_name. 
    par_command: str
        Shell command used to invoke parallelization. Eg. mpiexec, mpirun or on a Slurm system srun.
    modules: list
        List of strings for additional imports like for example:
        'from gpaw.utilities import h2gpts'
    **kwargs: 
        The kwargs will be interpreted as settings for the GPAW calculation, BUT, they MUST be specified as strings 
        rather than the Python objects that GPAW expects. This is because actually importing these modules from gpaw 
        somehow breaks os.command (and also subprocess.Popen). So for example if you want to use plane-wave mode, you 
        supply mode='PW(300)', NOT mode=PW(300)! 
        DO NOT LOAD ANYTHING FROM THE GPAW LIBRARY IN THE AGOX SCRIPT PARALLELIZATION WILL NOT WORK AS EXPECTED OR AT ALL.     

    The trickery with modules and **kwargs is handled using exec and eval, which is not the safest thing in the world 
    as it will execute all supplied string as Python code. So be careful with that!
    """

    implemented_properties = ['energy', 'forces']

    def __init__(self, parexe=True, par_command='mpiexec', script_name='gpaw_calc.py', settings_name='gpaw_settings.pckl', modules=[], **kwargs):
        Calculator.__init__(self)

        if importlib.util.find_spec('gpaw') is None:
            raise ModuleNotFoundError('GPAW Python module not found.')

        self.parexe = parexe
        self.par_command = par_command

        self.script_name = script_name
        self.settings_name = settings_name
    
        # All kwargs are assumed to be settings!
        self.settings_for_gpaw = kwargs
        self.settings_for_gpaw['modules'] = modules

        self.input_file = 'gpaw_io_input.traj'
        self.output_file = 'gpaw_io_output.traj'

        self.prepare()

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms=atoms, properties=properties, system_changes=system_changes)

        write(self.input_file, atoms)
        write(self.output_file, [])

        commandline = ''
        if self.parexe: 
            commandline += self.par_command + ' '

        commandline += 'python ' + self.script_name + ' ' + self.input_file + ' ' + self.output_file + ' ' + self.settings_name

        # Execute the command-line:
        #print(commandline)
        s = os.system(commandline)

        # Read the results:
        try:
            tnew = read(self.output_file)
        except Exception as e:
            self.error_handling(e)

        # Assign energies:
        self.results['energy'] = tnew.get_potential_energy(apply_constraint=False)
        try: # Not sure when this can happen? 
            self.results['forces'] = tnew.get_forces(apply_constraint=False)
        except Exception as e:
           self.error_handling(e)

        # Delete the files:
        os.remove(self.output_file)
        os.remove(self.input_file)        


    def error_handling(self, exception):
         # Catch the error and print it
        print('Error: ', exception)
        # Delete files to avoid possibility of errors. 
        os.remove(self.output_file)
        os.remove(self.input_file)
        # Raise an exception that is caught by the gauge.
        # Unlike the error produced by GPAW this should only be printed once so much easier to search for.. 
        raise Exception('Energy calculation failed')

    def prepare(self):
        # Write the pickle settings file:
        with open(self.settings_name, 'wb') as f:
            pickle.dump(self.settings_for_gpaw, f)

        # Write the script:
        with open(self.script_name, 'w') as f:
            print(calc_script, file=f)        



import numpy as np
from agox.environments.ABC_environment import EnvironmentBaseClass
from ase import Atoms
from ase.atoms import symbols2numbers
from ase.symbols import Symbols


class Environment(EnvironmentBaseClass):

    """
    Environment class for the generation of candidates with a single stoichoimetry.

    Parameters
    ----------
    template : Atoms
        The template structure to use.
    numbers : list, optional
        List of atomic numbers to add to the template, by default None.
    symbols : list, optional
        List of atomic symbols to add to the template, by default None.
    print_report : bool, optional
        If True, a report is printed, by default True.

    """

    def __init__(self, template, numbers=None, symbols=None, print_report=True,
                 **kwargs):
        
        super().__init__(**kwargs)

        # Both numbers and symbols cannot be specified:
        assert (numbers is not None) is not (symbols is not None) # XOR

        if numbers is not None:
            self._numbers = numbers
        elif symbols is not None:
            self._numbers = symbols2numbers(symbols)
        
        if type(template) is Atoms:
            template = self.convert_to_candidate_object(template)
        self._template = template

        if self.confinement_cell is None:
            print('No confinement cell specified. Using template cell.')
            self.confinement_cell = self._template.get_cell()
            self.confinement_corner = np.array([0, 0, 0])
        
        if print_report:
            self.environment_report()

    def get_template(self):
        return self._template.copy()

    def set_template(self, template):
        self._template = template

    def set_numbers(self, numbers):
        self._numbers = numbers

    def get_numbers(self):
        return self._numbers.copy()

    def get_missing_types(self):
        return np.sort(np.unique(self.get_numbers()))

    def get_all_types(self):
        return list(set(list(self._template.numbers) + list(self.get_numbers())))

    def get_identifier(self):
        return self.__hash__()

    def get_missing_indices(self):
        return np.arange(len(self._template), len(self._template)+len(self._numbers))

    def get_all_numbers(self):
        all_numbers = np.append(self.get_numbers(), self._template.get_atomic_numbers())
        return all_numbers

    def get_all_species(self):
        return list(Symbols(self.get_all_types()).species())

    def get_species(self):
        return list(np.unique(self.get_all_species()))

    def get_atoms(self):
        atoms = self.get_template()
        atoms += Atoms(self.get_numbers())
        return atoms
    
    def match(self, candidate):
        cand_numbers = candidate.get_atomic_numbers()
        env_numbers = self.get_all_numbers()

        stoi_match = (np.sort(cand_numbers) == np.sort(env_numbers)).all() # Not very efficient, but does it matter? Should pobably only use this function for debugging.
        template_match = (candidate.positions[0:len(candidate.template)] == self._template.positions).all()
        return stoi_match*template_match

    def __hash__(self):
        feature = tuple(self.get_numbers()) + tuple(self._template.get_atomic_numbers()) + tuple(self._template.get_positions().flatten().tolist())
        return hash(feature)    

    def environment_report(self):
        from agox.writer import pretty_print, header_print

        header_print('Environment Properties')
        tab = '    '

        missing_numbers = self.get_numbers()

        pretty_print('Atoms in search:')  
        for number in np.unique(missing_numbers):
            symbols_object = Symbols([number])
            specie = symbols_object.species().pop()
            count = np.count_nonzero(missing_numbers == number)
            pretty_print(tab + f'{specie} = {count}')

        total_symbols = Symbols(self.get_all_numbers())
        pretty_print(f'Template formula: {self._template.get_chemical_formula()}')
        pretty_print(f'Full formula: {total_symbols.get_chemical_formula()}')

        pretty_print('Cell:')
        for cell_vec in self._template.get_cell():
            pretty_print(tab + '{:4.2f} {:4.2f} {:4.2f}'.format(*cell_vec))
        pretty_print('Periodicity:')
        pretty_print(tab + '{} {} {}'.format(*self._template.pbc))

        pretty_print(f'Box constraint: {self.use_box_constraint}')
        if self.use_box_constraint:
            assert self.confinement_cell is not None
            assert self.confinement_corner is not None
            pretty_print('Confinement corner')
            pretty_print(tab + '{:4.2f} {:4.2f} {:4.2f}'.format(*self.confinement_corner))
            pretty_print('Confinement cell:')
            for cell_vec in self.confinement_cell:
                pretty_print(tab + '{:4.2f} {:4.2f} {:4.2f}'.format(*cell_vec))

        header_print('')
    
